"""Our log handlers for Python's logging package.
"""

from __future__ import division, absolute_import, print_function

import os
import socket
import time

import logging
import logging.handlers
from logging import LoggerAdapter

import skytools
import skytools.tnetstrings

try:
    unicode
except NameError:
    unicode = str   # noqa

__all__ = ['getLogger']

# add TRACE level
TRACE = 5
logging.TRACE = TRACE
logging.addLevelName(TRACE, 'TRACE')

# extra info to be added to each log record
_service_name = 'unknown_svc'
_job_name = 'unknown_job'
_hostname = socket.gethostname()
try:
    _hostaddr = socket.gethostbyname(_hostname)
except:
    _hostaddr = "0.0.0.0"
_log_extra = {
    'job_name': _job_name,
    'service_name': _service_name,
    'hostname': _hostname,
    'hostaddr': _hostaddr,
}
def set_service_name(service_name, job_name):
    """Set info about current script."""
    global _service_name, _job_name

    _service_name = service_name
    _job_name = job_name

    _log_extra['job_name'] = _job_name
    _log_extra['service_name'] = _service_name

#
# How to make extra fields available to all log records:
# 1. Use own getLogger()
#    - messages logged otherwise (eg. from some libs)
#      will crash the logging.
# 2. Fix record in own handlers
#    - works only with custom handlers, standard handlers will
#      crash is used with custom fmt string.
# 3. Change root logger
#    - can't do it after non-root loggers are initialized,
#      doing it before will depend on import order.
# 4. Update LogRecord.__dict__
#    - fails, as formatter uses obj.__dict__ directly.
# 5. Change LogRecord class
#    - ugly but seems to work.
#
_OldLogRecord = logging.LogRecord
class _NewLogRecord(_OldLogRecord):
    def __init__(self, *args):
        super(_NewLogRecord, self).__init__(*args)
        self.__dict__.update(_log_extra)
logging.LogRecord = _NewLogRecord


# configurable file logger
class EasyRotatingFileHandler(logging.handlers.RotatingFileHandler):
    """Easier setup for RotatingFileHandler."""
    def __init__(self, filename, maxBytes=10*1024*1024, backupCount=3):
        """Args same as for RotatingFileHandler, but in filename '~' is expanded."""
        fn = os.path.expanduser(filename)
        super(EasyRotatingFileHandler, self).__init__(fn, maxBytes=maxBytes, backupCount=backupCount)


# send JSON message over UDP
class UdpLogServerHandler(logging.handlers.DatagramHandler):
    """Sends log records over UDP to logserver in JSON format."""

    # map logging levels to logserver levels
    _level_map = {
        logging.DEBUG   : 'DEBUG',
        logging.INFO    : 'INFO',
        logging.WARNING : 'WARN',
        logging.ERROR   : 'ERROR',
        logging.CRITICAL: 'FATAL',
    }

    # JSON message template
    _log_template = '{\n\t'\
        '"logger": "skytools.UdpLogServer",\n\t'\
        '"timestamp": %.0f,\n\t'\
        '"level": "%s",\n\t'\
        '"thread": null,\n\t'\
        '"message": %s,\n\t'\
        '"properties": {"application":"%s", "apptype": "%s", "type": "sys", "hostname":"%s", "hostaddr": "%s"}\n'\
        '}\n'

    # cut longer msgs
    MAXMSG = 1024

    def makePickle(self, record):
        """Create message in JSON format."""
        # get & cut msg
        msg = self.format(record)
        if len(msg) > self.MAXMSG:
            msg = msg[:self.MAXMSG]
        txt_level = self._level_map.get(record.levelno, "ERROR")
        hostname = _hostname
        hostaddr = _hostaddr
        jobname = _job_name
        svcname = _service_name
        pkt = self._log_template % (time.time()*1000, txt_level, skytools.quote_json(msg),
                jobname, svcname, hostname, hostaddr)
        return pkt

    def send(self, s):
        """Disable socket caching."""
        sock = self.makeSocket()
        sock.sendto(s, (self.host, self.port))
        sock.close()


# send TNetStrings message over UDP
class UdpTNetStringsHandler(logging.handlers.DatagramHandler):
    """ Sends log records in TNetStrings format over UDP. """

    # LogRecord fields to send
    send_fields = [
        'created', 'exc_text', 'levelname', 'levelno', 'message', 'msecs', 'name',
        'hostaddr', 'hostname', 'job_name', 'service_name']

    _udp_reset = 0

    def makePickle(self, record):
        """ Create message in TNetStrings format.
        """
        msg = {}
        self.format(record) # render 'message' attribute and others
        for k in self.send_fields:
            msg[k] = record.__dict__[k]
        tnetstr = skytools.tnetstrings.dumps(msg)
        return tnetstr

    def send(self, s):
        """ Cache socket for a moment, then recreate it.
        """
        now = time.time()
        if now - 1 > self._udp_reset:
            if self.sock:
                self.sock.close()
            self.sock = self.makeSocket()
            self._udp_reset = now
        self.sock.sendto(s, (self.host, self.port))


class LogDBHandler(logging.handlers.SocketHandler):
    """Sends log records into PostgreSQL server.

    Additionally, does some statistics aggregating,
    to avoid overloading log server.

    It subclasses SocketHandler to get throtthling for
    failed connections.
    """

    # map codes to string
    _level_map = {
        logging.DEBUG   : 'DEBUG',
        logging.INFO    : 'INFO',
        logging.WARNING : 'WARNING',
        logging.ERROR   : 'ERROR',
        logging.CRITICAL: 'FATAL',
    }

    def __init__(self, connect_string):
        """
        Initializes the handler with a specific connection string.
        """

        super(LogDBHandler, self).__init__(None, None)
        self.closeOnError = 1

        self.connect_string = connect_string

        self.stat_cache = {}
        self.stat_flush_period = 60
        # send first stat line immediately
        self.last_stat_flush = 0

    def createSocket(self):
        try:
            super(LogDBHandler, self).createSocket()
        except:
            self.sock = self.makeSocket()

    def makeSocket(self, timeout=1):
        """Create server connection.
        In this case its not socket but database connection."""

        db = skytools.connect_database(self.connect_string)
        db.set_isolation_level(0) # autocommit
        return db

    def emit(self, record):
        """Process log record."""

        # we do not want log debug messages
        if record.levelno < logging.INFO:
            return

        try:
            self.process_rec(record)
        except (SystemExit, KeyboardInterrupt):
            raise
        except:
            self.handleError(record)

    def process_rec(self, record):
        """Aggregate stats if needed, and send to logdb."""
        # render msg
        msg = self.format(record)

        # dont want to send stats too ofter
        if record.levelno == logging.INFO and msg and msg[0] == "{":
            self.aggregate_stats(msg)
            if time.time() - self.last_stat_flush >= self.stat_flush_period:
                self.flush_stats(_job_name)
            return

        if record.levelno < logging.INFO:
            self.flush_stats(_job_name)

        # dont send more than one line
        ln = msg.find('\n')
        if ln > 0:
            msg = msg[:ln]

        txt_level = self._level_map.get(record.levelno, "ERROR")
        self.send_to_logdb(_job_name, txt_level, msg)

    def aggregate_stats(self, msg):
        """Sum stats together, to lessen load on logdb."""

        msg = msg[1:-1]
        for rec in msg.split(", "):
            k, v = rec.split(": ")
            agg = self.stat_cache.get(k, 0)
            if v.find('.') >= 0:
                agg += float(v)
            else:
                agg += int(v)
            self.stat_cache[k] = agg

    def flush_stats(self, service):
        """Send acquired stats to logdb."""
        res = []
        for k, v in self.stat_cache.items():
            res.append("%s: %s" % (k, str(v)))
        if len(res) > 0:
            logmsg = "{%s}" % ", ".join(res)
            self.send_to_logdb(service, "INFO", logmsg)
        self.stat_cache = {}
        self.last_stat_flush = time.time()

    def send_to_logdb(self, service, level, msg):
        """Actual sending is done here."""

        if self.sock is None:
            self.createSocket()

        if self.sock:
            logcur = self.sock.cursor()
            query = "select * from log.add(%s, %s, %s)"
            logcur.execute(query, [level, service, msg])


# fix unicode bug in SysLogHandler
class SysLogHandler(logging.handlers.SysLogHandler):
    """Fixes unicode bug in logging.handlers.SysLogHandler."""

    # be compatible with both 2.6 and 2.7
    socktype = socket.SOCK_DGRAM

    _udp_reset = 0

    def _custom_format(self, record):
        msg = self.format(record) + '\000'

        # We need to convert record level to lowercase, maybe this will
        # change in the future.
        prio = '<%d>' % self.encodePriority(self.facility,
                                            self.mapPriority(record.levelname))
        msg = prio + msg
        return msg

    def emit(self, record):
        """
        Emit a record.

        The record is formatted, and then sent to the syslog server. If
        exception information is present, it is NOT sent to the server.
        """
        msg = self._custom_format(record)
        # Message is a string. Convert to bytes as required by RFC 5424
        if isinstance(msg, unicode):
            msg = msg.encode('utf-8')
            ## this puts BOM in wrong place
            #if codecs:
            #    msg = codecs.BOM_UTF8 + msg
        try:
            if self.unixsocket:
                try:
                    self.socket.send(msg)
                except socket.error:
                    self._connect_unixsocket(self.address)
                    self.socket.send(msg)
            elif self.socktype == socket.SOCK_DGRAM:
                now = time.time()
                if now - 1 > self._udp_reset:
                    self.socket.close()
                    self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                    self._udp_reset = now
                self.socket.sendto(msg, self.address)
            else:
                self.socket.sendall(msg)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

class SysLogHostnameHandler(SysLogHandler):
    """Slightly modified standard SysLogHandler - sends also hostname and service type"""

    def _custom_format(self, record):
        msg = self.format(record)
        format_string = '<%d> %s %s %s\000'
        msg = format_string % (self.encodePriority(self.facility, self.mapPriority(record.levelname)),
                               _hostname, _service_name, msg)
        return msg


# add missing aliases (that are in Logger class)
if not hasattr(LoggerAdapter, 'fatal'):
    LoggerAdapter.fatal = LoggerAdapter.critical
if not hasattr(LoggerAdapter, 'warn'):
    LoggerAdapter.warn = LoggerAdapter.warning

class SkyLogger(LoggerAdapter):
    def __init__(self, logger, extra):
        super(SkyLogger, self).__init__(logger, extra)
        self.name = logger.name
    def trace(self, msg, *args, **kwargs):
        """Log 'msg % args' with severity 'TRACE'."""
        self.log(TRACE, msg, *args, **kwargs)
    def addHandler(self, hdlr):
        """Add the specified handler to this logger."""
        self.logger.addHandler(hdlr)
    def isEnabledFor(self, level):
        """See if the underlying logger is enabled for the specified level."""
        return self.logger.isEnabledFor(level)

def getLogger(name=None, **kwargs_extra):
    """Get logger with extra functionality.

    Adds additional log levels, and extra fields to log record.

    name - name for logging.getLogger()
    kwargs_extra - extra fields to add to log record
    """
    log = logging.getLogger(name)
    return SkyLogger(log, kwargs_extra)



"""Check if SQL keywords are up-to-date.
"""

from __future__ import division, absolute_import, print_function

import sys
import re
import os.path

import skytools.quoting

from nose.tools import *

locations = [
    "/opt/src/pgsql/postgresql/src/include/parser/kwlist.h",
    "~/src/pgsql/postgres/src/include/parser/kwlist.h",
    "~/src/pgsql/pg95/src/include/parser/kwlist.h",
    "~/src/pgsql/pg94/src/include/parser/kwlist.h",
    "~/src/pgsql/pg93/src/include/parser/kwlist.h",
    "~/src/pgsql/pg92/src/include/parser/kwlist.h",
    "~/src/pgsql/pg91/src/include/parser/kwlist.h",
    "~/src/pgsql/pg90/src/include/parser/kwlist.h",
    "~/src/pgsql/pg84/src/include/parser/kwlist.h",
    "~/src/pgsql/pg83/src/include/parser/kwlist.h",
    "/usr/include/postgresql/9.5/server/parser/kwlist.h",
    "/usr/include/postgresql/9.4/server/parser/kwlist.h",
    "/usr/include/postgresql/9.3/server/parser/kwlist.h",
    "/usr/include/postgresql/9.2/server/parser/kwlist.h",
    "/usr/include/postgresql/9.1/server/parser/kwlist.h",
]

def _load_kwlist(fn, full_map, cur_map):
    fn = os.path.expanduser(fn)
    if not os.path.isfile(fn):
        return
    data = open(fn, 'rt').read()
    rc = re.compile(r'PG_KEYWORD[(]"(.*)" , \s* \w+ , \s* (\w+) [)]', re.X)
    for kw, cat in rc.findall(data):
        full_map[kw] = cat
        if cat == 'UNRESERVED_KEYWORD':
            continue
        if cat == 'COL_NAME_KEYWORD':
            continue
        cur_map[kw] = cat

def test_kwcheck():
    """Compare keyword list in quoting.py to the one in postgres sources
    """

    kwset = set(skytools.quoting._ident_kwmap)
    full_map = {}           # all types from kwlist.h
    cur_map = {}            # only kwlist.h
    new_list = []           # missing from kwset
    obsolete_list = []      # in kwset, but not in cur_map

    for fn in locations:
        _load_kwlist(fn, full_map, cur_map)
    if not full_map:
        return

    for kw in sorted(cur_map):
        if kw not in kwset:
            new_list.append((kw, cur_map[kw]))
        kwset.add(kw)

    for k in sorted(kwset):
        if k not in full_map:
            # especially obsolete
            obsolete_list.append( (k, '!FULL') )
        elif k not in cur_map:
            # slightly obsolete
            obsolete_list.append( (k, '!CUR') )

    eq_(new_list, [])

    # here we need to keep older keywords around longer
    #eq_(obsolete_list, [])




"""Extra tests for quoting module.
"""

from __future__ import division, absolute_import, print_function

import sys, time
import skytools.psycopgwrapper
import skytools._cquoting
import skytools._pyquoting
from decimal import Decimal

from skytools.testing import ordered_dict
from nose.tools import *

# create a DictCursor row
class fake_cursor:
    index = ordered_dict({'id': 0, 'data': 1})
    description = ['x', 'x']
dbrow = skytools.psycopgwrapper._CompatRow(fake_cursor())
dbrow[0] = '123'
dbrow[1] = 'value'

def try_func(qfunc, data_list):
    for val, exp in data_list:
        got = qfunc(val)
        eq_(got, exp)

def try_catch(qfunc, data_list, exc):
    for d in data_list:
        assert_raises(exc, qfunc, d)

def test_quote_literal():
    sql_literal = [
        [None, "null"],
        ["", "''"],
        ["a'b", "'a''b'"],
        [r"a\'b", r"E'a\\''b'"],
        [1, "'1'"],
        [True, "'True'"],
        [Decimal(1), "'1'"],
        [u'qwe', "'qwe'"]
    ]
    try_func(skytools._cquoting.quote_literal, sql_literal)
    try_func(skytools._pyquoting.quote_literal, sql_literal)
    try_func(skytools.quote_literal, sql_literal)

qliterals_common = [
    (r"""null""", None),
    (r"""NULL""", None),
    (r"""123""", "123"),
    (r"""''""", r""""""),
    (r"""'a''b''c'""", r"""a'b'c"""),
    (r"""'foo'""", r"""foo"""),
    (r"""E'foo'""", r"""foo"""),
    (r"""E'a\n\t\a\b\0\z\'b'""", "a\n\t\x07\x08\x00z'b"),
    (r"""$$$$""", r""),
    (r"""$$qw$e$z$$""", r"qw$e$z"),
    (r"""$qq$$aa$$$'"\\$qq$""", '$aa$$$\'"\\\\'),
    (u"'qwe'", 'qwe'),
]

bad_dol_literals = [
    ('$$', '$$'),
    #('$$q', '$$q'),
    ('$$q$', '$$q$'),
    ('$q$q$', '$q$q$'),
    ('$q$q$x$', '$q$q$x$'),
]

def test_unquote_literal():
    qliterals_nonstd = qliterals_common + [
        (r"""'a\\b\\c'""", r"""a\b\c"""),
        (r"""e'a\\b\\c'""", r"""a\b\c"""),
    ]
    try_func(skytools._cquoting.unquote_literal, qliterals_nonstd)
    try_func(skytools._pyquoting.unquote_literal, qliterals_nonstd)
    try_func(skytools.unquote_literal, qliterals_nonstd)

    for v1, v2 in bad_dol_literals:
        assert_raises(ValueError, skytools._pyquoting.unquote_literal, v1)
        assert_raises(ValueError, skytools._cquoting.unquote_literal, v1)
        assert_raises(ValueError, skytools.unquote_literal, v1)

def test_unquote_literal_std():
    qliterals_std = qliterals_common + [
        (r"''", r""),
        (r"'foo'", r"foo"),
        (r"E'foo'", r"foo"),
        (r"'\\''z'", r"\\'z"),
    ]
    for val, exp in qliterals_std:
        eq_(skytools._cquoting.unquote_literal(val, True), exp)
        eq_(skytools._pyquoting.unquote_literal(val, True), exp)
        eq_(skytools.unquote_literal(val, True), exp)

def test_quote_copy():
    sql_copy = [
        [None, "\\N"],
        ["", ""],
        ["a'\tb", "a'\\tb"],
        [r"a\'b", r"a\\'b"],
        [1, "1"],
        [True, "True"],
        [u"qwe", "qwe"],
        [Decimal(1), "1"],
    ]
    try_func(skytools._cquoting.quote_copy, sql_copy)
    try_func(skytools._pyquoting.quote_copy, sql_copy)
    try_func(skytools.quote_copy, sql_copy)

def test_quote_bytea_raw():
    sql_bytea_raw = [
        [None, None],
        [b"", ""],
        [b"a'\tb", "a'\\011b"],
        [b"a\\'b", r"a\\'b"],
        [b"\t\344", r"\011\344"],
    ]
    try_func(skytools._cquoting.quote_bytea_raw, sql_bytea_raw)
    try_func(skytools._pyquoting.quote_bytea_raw, sql_bytea_raw)
    try_func(skytools.quote_bytea_raw, sql_bytea_raw)

def test_quote_bytea_raw_fail():
    assert_raises(TypeError, skytools._pyquoting.quote_bytea_raw, u'qwe')
    #assert_raises(TypeError, skytools._cquoting.quote_bytea_raw, u'qwe')
    #assert_raises(TypeError, skytools.quote_bytea_raw, 'qwe')

def test_quote_ident():
    sql_ident = [
        ['', '""'],
        ["a'\t\\\"b", '"a\'\t\\""b"'],
        ['abc_19', 'abc_19'],
        ['from', '"from"'],
        ['0foo', '"0foo"'],
        ['mixCase', '"mixCase"'],
        [u'utf', 'utf'],
    ]
    try_func(skytools.quote_ident, sql_ident)

def _sort_urlenc(func):
    def wrapper(data):
        res = func(data)
        return '&'.join(sorted(res.split('&')))
    return wrapper

def test_db_urlencode():
    t_urlenc = [
        [{}, ""],
        [{'a': 1}, "a=1"],
        [{'a': None}, "a"],
        [{'qwe': 1, u'zz': u"qwe"}, 'qwe=1&zz=qwe'],
        [ordered_dict({'qwe': 1, u'zz': u"qwe"}), 'qwe=1&zz=qwe'],
        [{'a': '\000%&'}, "a=%00%25%26"],
        [dbrow, 'data=value&id=123'],
        [{'a': Decimal("1")}, "a=1"],
    ]
    try_func(_sort_urlenc(skytools._cquoting.db_urlencode), t_urlenc)
    try_func(_sort_urlenc(skytools._pyquoting.db_urlencode), t_urlenc)
    try_func(_sort_urlenc(skytools.db_urlencode), t_urlenc)

def test_db_urldecode():
    t_urldec = [
        ["", {}],
        ["a=b&c", {'a': 'b', 'c': None}],
        ["&&b=f&&", {'b': 'f'}],
        [u"abc=qwe", {'abc': 'qwe'}],
        ["b=", {'b': ''}],
        ["b=%00%45", {'b': '\x00E'}],
    ]
    try_func(skytools._cquoting.db_urldecode, t_urldec)
    try_func(skytools._pyquoting.db_urldecode, t_urldec)
    try_func(skytools.db_urldecode, t_urldec)

def test_unescape():
    t_unesc = [
        ["", ""],
        ["\\N", "N"],
        ["abc", "abc"],
        [u"abc", "abc"],
        [r"\0\000\001\01\1", "\0\000\001\001\001"],
        [r"a\001b\tc\r\n", "a\001b\tc\r\n"],
    ]
    try_func(skytools._cquoting.unescape, t_unesc)
    try_func(skytools._pyquoting.unescape, t_unesc)
    try_func(skytools.unescape, t_unesc)

def test_quote_bytea_literal():
    bytea_raw = [
        [None, "null"],
        [b"", "''"],
        [b"a'\tb", "E'a''\\\\011b'"],
        [b"a\\'b", r"E'a\\\\''b'"],
        [b"\t\344", r"E'\\011\\344'"],
    ]
    try_func(skytools.quote_bytea_literal, bytea_raw)

def test_quote_bytea_copy():
    bytea_raw = [
        [None, "\\N"],
        [b"", ""],
        [b"a'\tb", "a'\\\\011b"],
        [b"a\\'b", r"a\\\\'b"],
        [b"\t\344", r"\\011\\344"],
    ]
    try_func(skytools.quote_bytea_copy, bytea_raw)

def test_quote_statement():
    sql = "set a=%s, b=%s, c=%s"
    args = [None, u"qwe'qwe", 6.6]
    eq_(skytools.quote_statement(sql, args), "set a=null, b='qwe''qwe', c='6.6'")

    sql = "set a=%(a)s, b=%(b)s, c=%(c)s"
    args = dict(a=None, b="qwe'qwe", c=6.6)
    eq_(skytools.quote_statement(sql, args), "set a=null, b='qwe''qwe', c='6.6'")

def test_quote_json():
    json_string_vals = [
        [None, "null"],
        ['', '""'],
        [u'xx', '"xx"'],
        ['qwe"qwe\t', '"qwe\\"qwe\\t"'],
        ['\x01', '"\\u0001"'],
    ]
    try_func(skytools.quote_json, json_string_vals)

def test_unquote_ident():
    idents = [
        ['qwe', 'qwe'],
        [u'qwe', 'qwe'],
        ['"qwe"', 'qwe'],
        ['"q""w\\\\e"', 'q"w\\\\e'],
    ]
    try_func(skytools.unquote_ident, idents)

@raises(Exception)
def test_unquote_ident_fail():
    skytools.unquote_ident('asd"asd')






import os.path

import io

from nose.tools import *

from skytools.config import (Config, NoOptionError, NoSectionError,
        ConfigError, InterpolationError,
        ExtendedConfigParser, ExtendedCompatConfigParser)

TOP = os.path.dirname(__file__)
CONFIG = os.path.join(TOP, 'config.ini')

def test_config_str():
    cf = Config('base', CONFIG)
    eq_(cf.get('foo'), '1')
    eq_(cf.get('missing', 'q'), 'q')
    assert_raises(NoOptionError, cf.get, 'missing')

def test_config_int():
    cf = Config('base', CONFIG)
    eq_(cf.getint('foo'), 1)
    eq_(cf.getint('missing', 2), 2)
    assert_raises(NoOptionError, cf.getint, 'missing')

def test_config_float():
    cf = Config('base', CONFIG)
    eq_(cf.getfloat('float-val'), 2.0)
    eq_(cf.getfloat('missing', 3.0), 3.0)
    assert_raises(NoOptionError, cf.getfloat, 'missing')

def test_config_bool():
    cf = Config('base', CONFIG)
    eq_(cf.getboolean('bool-true1'), True)
    eq_(cf.getboolean('bool-true2'), True)
    eq_(cf.getboolean('missing', True), True)
    assert_raises(NoOptionError, cf.getboolean, 'missing')

    eq_(cf.getboolean('bool-false1'), False)
    eq_(cf.getboolean('bool-false2'), False)
    eq_(cf.getboolean('missing', False), False)
    assert_raises(NoOptionError, cf.getbool, 'missing')

def test_config_list():
    cf = Config('base', CONFIG)
    eq_(cf.getlist('list-val1'), [])
    eq_(cf.getlist('list-val2'), ['a', '1', 'asd', 'ppp'])
    eq_(cf.getlist('missing', [1]), [1])
    assert_raises(NoOptionError, cf.getlist, 'missing')

def test_config_dict():
    cf = Config('base', CONFIG)
    eq_(cf.getdict('dict-val1'), {})
    eq_(cf.getdict('dict-val2'), {'a': '1', 'b': '2', 'z': 'z'})
    eq_(cf.getdict('missing', {'a':1}), {'a':1})
    assert_raises(NoOptionError, cf.getdict, 'missing')

def test_config_file():
    cf = Config('base', CONFIG)
    eq_(cf.getfile('file-val1'), '-')
    eq_(cf.getfile('file-val2')[0], '/')
    eq_(cf.getfile('missing', 'qwe'), 'qwe')
    assert_raises(NoOptionError, cf.getfile, 'missing')

def test_config_bytes():
    cf = Config('base', CONFIG)
    eq_(cf.getbytes('bytes-val1'), 4)
    eq_(cf.getbytes('bytes-val2'), 2048)
    eq_(cf.getbytes('missing', '3k'), 3072)
    assert_raises(NoOptionError, cf.getbytes, 'missing')

def test_config_wildcard():
    cf = Config('base', CONFIG)

    eq_(cf.get_wildcard('wild-*-*', ['a', 'b']), 'w.a.b')
    eq_(cf.get_wildcard('wild-*-*', ['a', 'x']), 'w.a')
    eq_(cf.get_wildcard('wild-*-*', ['q', 'b']), 'w2')
    eq_(cf.get_wildcard('missing-*-*', ['1', '2'], 'def'), 'def')
    assert_raises(NoOptionError, cf.get_wildcard, 'missing-*-*', ['1', '2'])

def test_config_default():
    cf = Config('base', CONFIG)
    eq_(cf.get('all'), 'yes')

def test_config_other():
    cf = Config('base', CONFIG)
    eq_(sorted(cf.sections()), ['base', 'other'])
    assert_true(cf.has_section('base'))
    assert_true(cf.has_section('other'))
    assert_false(cf.has_section('missing'))
    assert_false(cf.has_section('DEFAULT'))

    assert_false(cf.has_option('missing'))
    assert_true(cf.has_option('all'))
    assert_true(cf.has_option('foo'))

    cf2 = cf.clone('other')
    eq_(sorted(cf2.options()), ['all', 'config_dir', 'config_file',
        'host_name', 'job_name', 'service_name', 'test'])
    eq_(len(cf2.items()), len(cf2.options()))

def test_loading():
    assert_raises(NoSectionError, Config, 'random', CONFIG)
    assert_raises(ConfigError, Config, 'random', 'random.ini')

def test_nofile():
    cf = Config('base', None, user_defs = {'a': '1'})
    eq_(cf.sections(), ['base'])
    eq_(cf.get('a'), '1')

    cf = Config('base', None, user_defs = {'a': '1'}, ignore_defs=True)
    eq_(cf.get('a', '2'), '2')

def test_override():
    cf = Config('base', CONFIG, override = {'foo': 'overrided'})
    eq_(cf.get('foo'), 'overrided')

def test_vars():
    cf = Config('base', CONFIG)
    eq_(cf.get('vars1'), 'V2=V3=Q3')

    assert_raises(InterpolationError, cf.get, 'bad1')


def test_extended_compat():
    config = u'[foo]\nkey = ${sub} $${nosub}\nsub = 2\n[bar]\nkey = ${foo:key}\n'
    cf = ExtendedCompatConfigParser()
    cf.readfp(io.StringIO(config), 'conf.ini')
    eq_(cf.get('bar', 'key'), '2 ${nosub}')

    config = u'[foo]\nloop1= ${loop1}\nloop2 = ${loop3}\nloop3 = ${loop2}\n'
    cf = ExtendedCompatConfigParser()
    cf.readfp(io.StringIO(config), 'conf.ini')
    assert_raises(InterpolationError, cf.get, 'foo', 'loop1')
    assert_raises(InterpolationError, cf.get, 'foo', 'loop2')

    config = u'[foo]\nkey = %(sub)s ${sub}\nsub = 2\n[bar]\nkey = %(foo:key)s\nkey2 = ${foo:key}\n'
    cf = ExtendedCompatConfigParser()
    cf.readfp(io.StringIO(config), 'conf.ini')
    eq_(cf.get('bar', 'key2'), '2 2')
    assert_raises(NoOptionError, cf.get, 'bar', 'key')

    config = u'[foo]\nkey = ${bad:xxx}\n[bad]\nsub = 1\n'
    cf = ExtendedCompatConfigParser(); cf.readfp(io.StringIO(config), 'conf.ini')
    assert_raises(NoOptionError, cf.get, 'foo', 'key')





"""Atomic append of gzipped data.

The point is - if several gzip streams are concatenated,
they are read back as one whole stream.
"""

from __future__ import division, absolute_import, print_function

import gzip
from io import BytesIO

__all__ = ['gzip_append']

#
# gzip storage
#
def gzip_append(filename, data, level=6):
    """Append a block of data to file with safety checks."""

    # compress data
    buf = BytesIO()
    g = gzip.GzipFile(fileobj=buf, compresslevel=level, mode="w")
    g.write(data)
    g.close()
    zdata = buf.getvalue()

    # append, safely
    f = open(filename, "ab+", 0)
    f.seek(0, 2)
    pos = f.tell()
    try:
        f.write(zdata)
        f.close()
    except Exception as ex:
        # rollback on error
        f.seek(pos, 0)
        f.truncate()
        f.close()
        raise ex



"""Setup for skytools module.
"""

from setuptools import setup, Extension

import sys

# don't build C module on win32 as it's unlikely to have dev env
BUILD_C_MOD = 1
if sys.platform == 'win32':
    BUILD_C_MOD = 0

# check if building C is allowed
c_modules = []
if BUILD_C_MOD:
    c_modules = [
        Extension("skytools._cquoting", ['modules/cquoting.c']),
        Extension("skytools._chashtext", ['modules/hashtext.c']),
    ]

# run actual setup
setup(
    name = "skytools",
    license = "ISC",
    version = '3.3',
    url = "https://github.com/pgq/python-skytools",
    maintainer = "Marko Kreen",
    maintainer_email = "markokr@gmail.com",
    description = "Utilities for database scripts",
    packages = ['skytools'],
    ext_modules = c_modules,
    zip_safe = False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Database",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
)



"""
Implementation of Postgres hashing function.

hashtext_old() - used up to PostgreSQL 8.3
hashtext_new() - used since PostgreSQL 8.4

>>> import skytools._chashtext
>>> for i in range(3):
...     print([hashtext_new_py(b'x' * (i*5 + j)) for j in range(5)])
[-1477818771, 1074944137, -1086392228, -1992236649, -1379736791]
[-370454118, 1489915569, -66683019, -2126973000, 1651296771]
[755764456, -1494243903, 631527812, 28686851, -9498641]
>>> for i in range(3):
...     print([hashtext_old_py(b'x' * (i*5 + j)) for j in range(5)])
[-863449762, 37835117, 294739542, -320432768, 1007638138]
[1422906842, -261065348, 59863994, -162804943, 1736144510]
[-682756517, 317827663, -495599455, -1411793989, 1739997714]
>>> data = b'HypficUjFitraxlumCitcemkiOkIkthi'
>>> p = [hashtext_old_py(data[:l]) for l in range(len(data)+1)]
>>> c = [hashtext_old(data[:l]) for l in range(len(data)+1)]
>>> assert p == c, '%s <> %s' % (p, c)
>>> p == c
True
>>> p = [hashtext_new_py(data[:l]) for l in range(len(data)+1)]
>>> c = [hashtext_new(data[:l]) for l in range(len(data)+1)]
>>> assert p == c, '%s <> %s' % (p, c)
>>> p == c
True
"""

from __future__ import division, absolute_import, print_function

import sys
import struct

try:
    from skytools._chashtext import hashtext_old, hashtext_new
except ImportError:
    def hashtext_old(v):
        return hashtext_old_py(v)
    def hashtext_new(v):
        return hashtext_new_py(v)

__all__ = ["hashtext_old", "hashtext_new"]

# pad for last partial block
PADDING = b'\0' * 12

def uint32(x):
    """python does not have 32 bit integer so we need this hack to produce uint32 after bit operations"""
    return x & 0xffffffff

#
# Old Postgres hashtext() - lookup2 with custom initval
#

FMT_OLD = struct.Struct("<LLL")

def mix_old(a, b, c):
    c = uint32(c)

    a -= b; a -= c; a = uint32(a ^ (c>>13))
    b -= c; b -= a; b = uint32(b ^ (a<<8))
    c -= a; c -= b; c = uint32(c ^ (b>>13))
    a -= b; a -= c; a = uint32(a ^ (c>>12))
    b -= c; b -= a; b = uint32(b ^ (a<<16))
    c -= a; c -= b; c = uint32(c ^ (b>>5))
    a -= b; a -= c; a = uint32(a ^ (c>>3))
    b -= c; b -= a; b = uint32(b ^ (a<<10))
    c -= a; c -= b; c = uint32(c ^ (b>>15))

    return a, b, c

def hashtext_old_py(k):
    """Old Postgres hashtext()"""

    remain = len(k)
    pos = 0
    a = b = 0x9e3779b9
    c = 3923095

    # handle most of the key
    while remain >= 12:
        a2, b2, c2 = FMT_OLD.unpack_from(k, pos)
        a, b, c = mix_old(a + a2, b + b2, c + c2)
        pos += 12
        remain -= 12

    # handle the last 11 bytes
    a2, b2, c2 = FMT_OLD.unpack_from(k[pos:] + PADDING, 0)

    # the lowest byte of c is reserved for the length
    c2 = (c2 << 8) + len(k)

    a, b, c = mix_old(a + a2, b + b2, c + c2)

    # convert to signed int
    if c & 0x80000000:
        c = -0x100000000 + c

    return int(c)

#
# New Postgres hashtext() - hacked lookup3:
# - custom initval
# - calls mix() when len=12
# - shifted c in last block on little-endian
#

FMT_NEW = struct.Struct("=LLL")

def rol32(x, k):
    return ((x)<<(k)) | (uint32(x)>>(32-(k)))

def mix_new(a, b, c):
    a -= c;  a ^= rol32(c, 4);  c += b
    b -= a;  b ^= rol32(a, 6);  a += c
    c -= b;  c ^= rol32(b, 8);  b += a
    a -= c;  a ^= rol32(c, 16); c += b
    b -= a;  b ^= rol32(a, 19); a += c
    c -= b;  c ^= rol32(b, 4);  b += a

    return uint32(a), uint32(b), uint32(c)

def final_new(a, b, c):
    c ^= b; c -= rol32(b, 14)
    a ^= c; a -= rol32(c, 11)
    b ^= a; b -= rol32(a, 25)
    c ^= b; c -= rol32(b, 16)
    a ^= c; a -= rol32(c, 4)
    b ^= a; b -= rol32(a, 14)
    c ^= b; c -= rol32(b, 24)

    return uint32(a), uint32(b), uint32(c)

def hashtext_new_py(k):
    """New Postgres hashtext()"""
    remain = len(k)
    pos = 0
    a = b = c = 0x9e3779b9 + len(k) + 3923095

    # handle most of the key
    while remain >= 12:
        a2, b2, c2 = FMT_NEW.unpack_from(k, pos)
        a, b, c = mix_new(a + a2, b + b2, c + c2)
        pos += 12
        remain -= 12

    # handle the last 11 bytes
    a2, b2, c2 = FMT_NEW.unpack_from(k[pos:] + PADDING, 0)
    if sys.byteorder == 'little':
        c2 = c2 << 8
    a, b, c = final_new(a + a2, b + b2, c + c2)

    # convert to signed int
    if c & 0x80000000:
        c = -0x100000000 + c

    return int(c)




"""TNetStrings.

>>> def ustr(v): return repr(v).replace("u'", "'")
>>> def nstr(b): return b.decode('utf8')
>>> if isinstance('qwe', bytes): nstr = str
>>> vals = (None, False, True, 222, 333.0, "foo", u"bar")
>>> tnvals = [nstr(dumps(v)) for v in vals]
>>> tnvals
['0:~', '5:false!', '4:true!', '3:222#', '5:333.0^', '3:foo,', '3:bar,']
>>> vals2 = [[], (), {}, [1,2], {'a':'b'}]
>>> tnvals2 = [nstr(dumps(v)) for v in vals2]
>>> tnvals2
['0:]', '0:]', '0:}', '8:1:1#1:2#]', '8:1:a,1:b,}']
>>> ustr([parse(dumps(v)) for v in vals])
"[None, False, True, 222, 333.0, 'foo', 'bar']"
>>> ustr([parse(dumps(v)) for v in vals2])
"[[], [], {}, [1, 2], {'a': 'b'}]"
>>> nstr(dumps([memoryview(b'zzz'),b'qqq']))
'12:3:zzz,3:qqq,]'

Error handling:

>>> def errtest(bstr, exc):
...     try:
...         loads(bstr)
...         raise Exception('no exception')
...     except exc:
...         return None
>>> errtest(b'4:qwez!', ValueError)
>>> errtest(b'4:', ValueError)
>>> errtest(b'4:qwez', ValueError)
>>> errtest(b'4', ValueError)
>>> errtest(b'', ValueError)
>>> errtest(b'999999999999999999:z,', ValueError)
>>> errtest(u'qweqwe', TypeError)
>>> errtest(b'4:true!0:~', ValueError)
>>> errtest(b'1:X~', ValueError)
>>> errtest(b'8:1:1#1:2#}', ValueError)
>>> errtest(b'1:Xz', ValueError)

>>> dumps(divmod)
Traceback (most recent call last):
    ...
TypeError: Object type not supported: <built-in function divmod>



"""

from __future__ import division, absolute_import, print_function

import codecs

__all__ = ['loads', 'dumps']

try:
    unicode
except NameError:
    unicode = str   # noqa
    long = int      # noqa

_memstr_types = (unicode, bytes, memoryview)
_struct_types = (list, tuple, dict)
_inttypes = (int, long)

def _dumps(dst, val):
    if isinstance(val, _struct_types):
        tlenpos = len(dst)
        tlen = 0
        dst.append(None)
        if isinstance(val, dict):
            for k in val:
                tlen += _dumps(dst, k)
                tlen += _dumps(dst, val[k])
            dst.append(b'}')
        else:
            for v in val:
                tlen += _dumps(dst, v)
            dst.append(b']')
        dst[tlenpos] = b'%d:' % tlen
        return len(dst[tlenpos]) + tlen + 1
    elif isinstance(val, _memstr_types):
        if isinstance(val, unicode):
            bval = val.encode('utf8')
        elif isinstance(val, memoryview):
            bval = val.tobytes()
        else:
            bval = val
        tval = b'%d:%s,' % (len(bval), bval)
    elif isinstance(val, bool):
        tval = val and b'4:true!' or b'5:false!'
    elif isinstance(val, _inttypes):
        bval = b'%d' % val
        tval = b'%d:%s#' % (len(bval), bval)
    elif isinstance(val, float):
        bval = b'%r' % val
        tval = b'%d:%s^' % (len(bval), bval)
    elif val is None:
        tval = b'0:~'
    else:
        raise TypeError("Object type not supported: %r" % val)
    dst.append(tval)
    return len(tval)

_decode_utf8 = codecs.getdecoder('utf8')
def _loads(buf):
    pos = 0
    maxlen = min(len(buf), 9)
    while buf[pos:pos+1] != b':':
        pos += 1
        if pos > maxlen:
            raise ValueError("Too large length")
    lenbytes = buf[ : pos].tobytes()
    tlen = int(lenbytes)
    ofs = len(lenbytes) + 1
    endofs = ofs + tlen

    val = buf[ofs : endofs]
    code = buf[endofs : endofs + 1]
    rest = buf[endofs + 1:]
    if len(val) + 1 != tlen + len(code):
        raise ValueError("failed to load value, invalid length")

    if code == b',':
        return _decode_utf8(val)[0], rest
    elif code == b'#':
        return int(val.tobytes(), 10), rest
    elif code == b'^':
        return float(val.tobytes()), rest
    elif code == b']':
        listobj = []
        while val:
            elem, val = _loads(val)
            listobj.append(elem)
        return listobj, rest
    elif code == b'}':
        dictobj = {}
        while val:
            k, val = _loads(val)
            if not isinstance(k, unicode):
                raise ValueError("failed to load value, invalid key type")
            dictobj[k], val = _loads(val)
        return dictobj, rest
    elif code == b'!':
        if val == b'true':
            return True, rest
        if val == b'false':
            return False, rest
        raise ValueError("failed to load value, invalid boolean value")
    elif code == b'~':
        if val == b'':
            return None, rest
        raise ValueError("failed to load value, invalid null value")
    else:
        raise ValueError("failed to load value, invalid value code")

#
# Public API
#

def dumps(val):
    """Dump object tree as TNetString value.
    """
    dst = []
    _dumps(dst, val)
    return b''.join(dst)

def loads(binval):
    """Parse TNetstring from byte string.
    """
    if not isinstance(binval, (bytes, memoryview)):
        raise TypeError("Bytes or memoryview required")
    obj, rest = _loads(memoryview(binval))
    if rest:
        raise ValueError("Not all data processed")
    return obj

# old compat?
parse = loads
dump = dumps

